/* Cg_pointz: a remake of the simple waving points cube with nVidia's Cg
   Copyright (C) 2005 Angelo "Encelo" Theodorou
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifndef M_PI
    #define M_PI 3.14159265358979323846
#endif

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#include "textfile.h"

#define NPOINTS 50 /* Grid density */
#define DISTANCE 0.1f /* Space between points */
#define POINTSIZE 1.0f
#define NPLANES 10 /* Number of planes */
#define SIN_ANGLE 0.3f /* Sin waving speed */
#define ROTY 0.05f /* Y axis rotation speed */

int FrameTiming(void);
GLboolean CheckExtension(char *extName);
void ErrorCallback();


int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	int quit = 0;
	
	const SDL_VideoInfo* info = NULL;
	int width = 640;
	int height = 480;
	int bpp = 0;
	int flags = 0;

	Uint32 interval;
	GLuint Plane;
	int i, j, k;
	GLfloat angle = 0.0f;
	GLfloat roty = 0.0f;
	GLfloat points[NPOINTS][NPOINTS][3];
	GLenum errCode;
	const GLubyte *errString;

	
/* ----- SDL init --------------- */
	if(SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);
	
	info = SDL_GetVideoInfo();
	bpp = info->vfmt->BitsPerPixel;

	
/* ----- OpenGL attribute setting via SDL --------------- */
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	
/*	flags = SDL_OPENGL | SDL_FULLSCREEN; */
	flags = SDL_OPENGL;

	
/* ----- Setting up the screen surface --------------- */
	if((screen = SDL_SetVideoMode(width, height, bpp, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("Cg_pointz by Encelo", NULL);


/* ----- Checking for GL extensions --------------- */
	if ((CheckExtension("GL_ARB_vertex_program") &&
			CheckExtension("GL_ARB_fragment_program")) == GL_TRUE)
		printf("Ready for Cg vertex and fragment programs.\n");
	else {
		printf("Missing extensions for Cg vertex and fragment programs!\n");
		exit(-1);
	}


/* ----- OpenGL init --------------- */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

//	glEnable(GL_POINT_SMOOTH); /* It KILLS fragment shader performance */
//	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
	glPointSize(POINTSIZE);


/* ----- Shaders setup --------------- */
	CGcontext context;
	CGprogram vprogram, fprogram;
	CGprofile vprofile, fprofile;
	CGerror error;
	CGparameter modelViewMatrix, par1;

	context = cgCreateContext();

//	vprofile = CG_PROFILE_ARBVP1;
	vprofile = cgGLGetLatestProfile(CG_GL_VERTEX);
	cgGLSetOptimalOptions(vprofile);
	vprogram = cgCreateProgramFromFile(context, CG_SOURCE, "wavey_vs.cg", vprofile, "main", NULL);
	cgGLLoadProgram(vprogram);
	cgGLEnableProfile(vprofile);
	cgGLBindProgram(vprogram);
	
//	fprofile = CG_PROFILE_ARBFP1;
	fprofile = cgGLGetLatestProfile(CG_GL_FRAGMENT);	
	cgGLSetOptimalOptions(fprofile);
	fprogram = cgCreateProgramFromFile(context,	CG_SOURCE, "wavey_fs.cg", fprofile, "main", NULL);
	cgGLLoadProgram(fprogram);
	cgGLEnableProfile(fprofile);
	cgGLBindProgram(fprogram);

	modelViewMatrix = cgGetNamedParameter(vprogram, "ModelViewMatrix");
	par1 = cgGetNamedParameter(vprogram, "angle");

	cgSetErrorCallback(ErrorCallback);
	

/* ----- Plane init --------------- */
	for(i=0; i<NPOINTS; i++) {
		for(j=0; j<NPOINTS; j++) {
			points[i][j][0] = NPOINTS*(-DISTANCE/2.0f) + i*DISTANCE;
			points[i][j][1] = 0.0f;
			points[i][j][2] = NPOINTS*(DISTANCE/2.0f) - j*DISTANCE;
		}
	}

	Plane = glGenLists (1);
	
	
/* ----- Event cycle --------------- */
	while (!quit) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_QUIT:
				quit = 1;
				break;
			case SDL_KEYDOWN:
				switch( event.key.keysym.sym ){
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = 1;
					break;
				case SDLK_F1:
					SDL_WM_ToggleFullScreen(screen);
					break;
				}
			}
		}

		interval = FrameTiming();

		
/* ----- Blitting on the screen --------------- */
		glClear(GL_COLOR_BUFFER_BIT);

		angle += SIN_ANGLE * interval;
		cgGLSetParameter1f(par1, angle);
		roty += ROTY * interval;

		glNewList(Plane, GL_COMPILE);
		glBegin(GL_POINTS);
			for(i=0; i<NPOINTS; i++)
				for(j=0; j<NPOINTS; j++) {
					glVertex3f(points[i][j][0],	points[i][j][1], points[i][j][2]);
				}
		glEnd();
		glEndList();

		glLoadIdentity();
		glTranslatef(0.0f, 2.75f, -9.0f);
		glRotatef(30.0f, 1.0f, 0.0f, 0.0f);
		glRotatef(roty, 0.0f, 1.0f, 0.0f);

		for (k=0; k<NPLANES; k++) {
			glTranslatef(0.0f, -(NPOINTS*DISTANCE)/NPLANES, 0.0f);
			cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
			glCallList(Plane);
		}

		SDL_GL_SwapBuffers();	
/*		SDL_Delay(20); */ /* Decomment this if you want 1/50th screen update */
	}

	
/* ----- Quitting --------------- */
	cgDestroyProgram(fprogram);
	cgDestroyProgram(vprogram);
	cgDestroyContext(context);

	glDeleteLists(Plane, 1);
	SDL_Quit();
	return 0;	
}


/* Search for extName in the extension string */
/* From the OpenGL 1.1 Programming Guide */
GLboolean CheckExtension(char *extName)
{
    char *p = (char *) glGetString(GL_EXTENSIONS);
    char *end;
    int extNameLen;

    extNameLen = strlen(extName);
    end = p + strlen(p);

    while (p < end) {
        int n = strcspn(p, " ");
        if ((extNameLen == n) && (strncmp(extName, p, n) == 0)) {
            return GL_TRUE;
        }
        p += (n + 1);
    }
    return GL_FALSE;
}


/* Calculate frame interval and print FPS each 5s */
int FrameTiming(void)
{
	Uint32 interval;
	static Uint32 current, last = 0, five = 0, nframes = 0;

	current = SDL_GetTicks();
	nframes++;

	if (current - five > 5*1000) {
		printf("%u frames in 5 seconds = %.1f FPS\n", nframes, (float)nframes/5.0f);
		nframes = 0;
		five = current;
	}
	
	interval = current - last;
	last = current;
	
	return interval;
}


/* Print a string whenever a Cg error occurs */
void ErrorCallback() {
	const char* errorString = cgGetErrorString(cgGetError());
	printf("%s\n", errorString);
}
